// this flag is used to disable all convolution multiplication
#define MULTIPLY_BY_CONVOLUTION

#include "defs.h"

#include <assert.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <float.h>

#include "RealEstimate.h"
#include "DataManager.h"
#include "RealFuncs.h"
#include "RealConstants.h"

namespace RealLib {


// ErrorEstimate implementation

ErrorEstimate::ErrorEstimate(const u32 man, const i32 exp)
: m_Man(man), m_Exp(exp)
{
}

// conversion
ErrorEstimate::ErrorEstimate(const LongFloat &src, RoundingMode round)
{
	switch (src.Kind()) {
	case LongFloat::Normal: {
		// convert mantissa to double to extract the 32 most significant bits
		i32 exp = src.Exponent() * 32;
		double man = src.MantissaAsDouble();
		int mexp;
		man = ldexp(frexp(fabs(man), &mexp), 32) + round;
		// offset the exponent with the double's exponent
		exp += mexp;

		u32 m = u32(man);

		// we might have overflowed
		if (m == 0) {
			m = 1<<31;
			++exp;
		}

		m_Exp = exp;
		m_Man = m;
		break; }
	case LongFloat::Zero:
		m_Man = 0;
		m_Exp = minusinf;
		break;
	default:
		m_Man = 0;
		m_Exp = plusinf;
		break;
	}
}

ErrorEstimate::ErrorEstimate(const double man)
{
	switch (_fpclass(man)) {
	case _FPCLASS_PD:
	case _FPCLASS_ND:
	case _FPCLASS_PN:
	case _FPCLASS_NN: {
		int mexp;
		// split mantissa and exponent
		m_Man = u32(ldexp(frexp(fabs(man), &mexp), 32) + 1);
		// correct for possible overflow
		if (m_Man == 0) {
			m_Man = 1<<31;
			m_Exp = mexp + 1;
		} else m_Exp = mexp;
		break; }
	case _FPCLASS_PZ:
	case _FPCLASS_NZ:
		m_Man = 0;
		m_Exp = minusinf;
		break;
	default:	// nan, inf
		m_Man = 0;
		m_Exp = plusinf;
		break;
	}
}

// helpers for ErrorEstimate

// add mantissas of ErrorEstimates, round up
bool DoEEManAdd(u32 &man, u32 full, u32 part, u32 start)
{
	if (start >= 32) man = ++full;
	else {
		// see if we need to round up
		if (part & ((1<<start)-1)) 
			part = (part >> start) + 1;
		else part = (part >> start);

		man = full + part;
	}
	return (man < full);	// tricky check for carry
}

// sub mantissas of ErrorEstimates, round down
i32 DoEEManSub(u32 &man, u32 full, u32 part, u32 start, i32 exp)
{
	// substract
	if (start >= 32) man = --full;
	else {
		if (start > 0)
			if (part & ((1<<start)-1)) 
				part = (part >> start) + 1;
			else part = (part >> start);

		man = full - part;
	}
	assert (man <= full);

	// normalize
	if (man == 0) return ErrorEstimate::minusinf;
	while ((man & (1<<31))==0) {
		--exp;
		man <<= 1;
	}
	return exp;
}

ErrorEstimate ErrorEstimate::operator + (const ErrorEstimate &rhs) const
{
	u32 man;
	i32 exp;
	bool carry;

	// handle special cases
	if (m_Exp == plusinf || rhs.m_Exp == minusinf) return *this;
	if (rhs.m_Exp == plusinf || m_Exp == minusinf) return rhs;

	// do addition
	if (m_Exp == rhs.m_Exp) {
		exp = m_Exp;
		man = m_Man + rhs.m_Man;
		carry = true;
	} else if (m_Exp > rhs.m_Exp) {
		exp = m_Exp;
		carry = DoEEManAdd(man, m_Man, rhs.m_Man, m_Exp - rhs.m_Exp);
	} else {
		exp = rhs.m_Exp;
		carry = DoEEManAdd(man, rhs.m_Man, m_Man, rhs.m_Exp - m_Exp);
	}

	// update if carry
	if (carry) {
		if (man & 1) ++man;
		man = (man >> 1) | (1 << 31);
		exp = exp + 1;
	}

	return ErrorEstimate(man, exp);
}

ErrorEstimate ErrorEstimate::operator - (const ErrorEstimate &rhs) const
{
	u32 man;
	i32 exp;

	// handle specials
	if (m_Exp == plusinf || rhs.m_Exp == minusinf) return *this;
	if (rhs.m_Exp == plusinf || m_Exp == minusinf) return rhs;

	// errors are always positive, a negative result would mean error
	assert(m_Exp >= rhs.m_Exp);
	exp = DoEEManSub(man, m_Man, rhs.m_Man, m_Exp - rhs.m_Exp, m_Exp);

	return ErrorEstimate(man, exp);
}

ErrorEstimate ErrorEstimate::operator * (const ErrorEstimate &rhs) const
{
	i64 e = i64(m_Exp) + i64(rhs.m_Exp) - 1;
	
	// handle overflow and special cases
	if (m_Exp == plusinf || rhs.m_Exp == plusinf || e >= plusinf) return ErrorEstimate(0, plusinf);
	if (m_Exp == minusinf || rhs.m_Exp == minusinf || e <= minusinf) return ErrorEstimate(0, minusinf);

	// multiply. the result will at least have 1 in 62nd position
	// at most 1 in 63rd
	u64 m = u64(m_Man) * u64(rhs.m_Man);
	// round up if necessary
	if (u32(m<<1)) m = (m >> 31) + 1;
	else m = (m >> 31);

	// make room for the 63rd bit if it is not 0
	if (m>>32) {
		if (m & 1) ++m;
		m >>= 1;
		++e;
	}

	return ErrorEstimate(u32(m), i32(e));
}

ErrorEstimate ErrorEstimate::operator << (i32 howmuch) const
{ 
	// simply add to exponent saturating
	i64 e = i64(m_Exp) + i64(howmuch);
	
	if (m_Exp == plusinf || e >= plusinf) return ErrorEstimate(0, plusinf);
	if (m_Exp == minusinf || e <= minusinf) return ErrorEstimate(0, minusinf);

	return ErrorEstimate(m_Man, i32(e));
}


ErrorEstimate ErrorEstimate::recip() const
{
	if (m_Exp == plusinf) return ErrorEstimate(0, minusinf);
	if (m_Exp <= minusinf + 2) return ErrorEstimate(0, plusinf);

	// calculate
	i32 exp = -(m_Exp - 2);
	u32 man = u32(((u64(1) << 62) + m_Man - 1) / m_Man);

	// normalize
	if (!(man & (1<<31))) {
		man = (man << 1) + 1;
		--exp;
	}

	return ErrorEstimate(man, exp);
}

ErrorEstimate& ErrorEstimate::operator ++ ()
{
	if (++m_Man == 0) {
		m_Man = 1<<31;
		++m_Exp;
	}
	return *this;
}

double ErrorEstimate::AsDouble() const
{
	// this can very easily over or underflow
	return ldexp(m_Exp == plusinf ? 1.0 : m_Man, m_Exp - 32);
}

LongFloat ErrorEstimate::AsLongFloat() const
{
	// m_Man is u32, so we can't use LongFloat's i32 constructor
	// go through double, but separate the exponent to a remainder
	// which can safely be applied to a double, and a quotient
	// which can very quickly be added to a LongFloat's exponent.
	i32 expq = m_Exp >> 5;
	i32 expr = m_Exp - (expq << 5);

	return LongFloat(ldexp(m_Man, expr)).AddToExponent(expq);
}

double ErrorEstimate::DecimalsCorrect() const
{ 
	// returns the rough number of correct decimals
	return -double(m_Exp) / LOG_2_10; 
}


bool ErrorEstimate::operator >= (const ErrorEstimate &rhs) const
{
	if (m_Exp > rhs.m_Exp) return true;
	else if (m_Exp == rhs.m_Exp && 
		(m_Exp == plusinf || m_Exp == minusinf || m_Man >= rhs.m_Man))
		return true;
	else return false;
}

ErrorEstimate max(const ErrorEstimate &a, const ErrorEstimate &b) 
{
	return a >= b ? a : b;
}

ErrorEstimate min(const ErrorEstimate &a, const ErrorEstimate &b) 
{
	return a >= b ? b : a;
}

ErrorEstimate RoundingError()
{
	return ErrorEstimate(1<<31, -(g_WorkingPrecision - 1) * 32 - 1);
}


// Estimate implementation

Estimate::Estimate(const LongFloat &val, const ErrorEstimate &err)
: m_Value(val), m_Error(err)
{
	if (val.Kind() == LongFloat::Zero) {
		m_Value = m_Error.AsLongFloat();
		m_Error = ErrorEstimate(2.0);
	}
}

char* Estimate::AsDecimal(char *bufptr, u32 buflen, u32 prec)
{
	// must at least accomodate exponent
	assert (buflen >= 10);
	char *buffer = bufptr;

	// the format chosen is "[-].<mantissa>e<+/-><exponent>"
	// where mantissa has a leading non-zero decimal

	if (m_Value.IsNegative()) {
		buffer++[0] = '-';
		--buflen;
	}

	// handle special values
	switch (m_Value.Kind()) {
	case LongFloat::Nan:
		strcpy(buffer, "NaN");
		return bufptr;
	case LongFloat::Infinity:
		strcpy(buffer, "Infinity");
		return bufptr;
	case LongFloat::Zero:
		strcpy(buffer, "Zero");
		return bufptr;
	}
	if (m_Error.gePow2(0)) {
		strcpy(buffer, "Zero");
		return bufptr;
	}

	// calculate exponent: the least power of 10 that is greater than
	// or equal to the value. Using decimal logarithm and its ceiling.
	Estimate pwr = ln(abs(*this, prec), prec) / Constants.ln10.GetEstimate(prec);
	// the following operation is safe: the exponent is at most
	// 2^(32 * log(10, 2)) which fits in less than the 53 mantissa bits
	// of double
	int e = int(ceil(pwr.Value().AsDouble()));

	// GCC doesn't understand _itoa, use sprintf instead
	sprintf(buffer, "%+d", e);
	// now we know the length of the exponent. move it to the end of
	// the string
	int explen = strlen(buffer);
	strcpy(buffer + buflen - explen - 1, buffer);
	buffer[0] = '.';

	// divide the value by the exponent to form decimal mantissa
	Estimate div(pow(LongFloat(10), e));
	Estimate man = *this / div;

	// when the value is power of ten, we can get two possible
	// representations of the mantissa: 0.(9) and 1.(0). This
	// behavior is not an error as it is consistent with the
	// theory. The function used to convert the mantissa will
	// not display 1.(0) correctly. Thus we must handle the
	// case differently: just pretend it's 0.(9)
	if (man.Value().Exponent() > 0) {
		for (u32 i=1;i<buflen - explen - 2; ++i)
			buffer[i] = '9';
	} else 
		// convert it to decimal using LongFloat's function
		man.Value().MantissaAsDecimal(buffer+1, buflen - explen - 2);
	
	char *ptr = buffer + (buflen - explen - 2);
	*ptr = 'e';

	return bufptr;
}


// operations
Estimate Estimate::operator - () const
{
	return Estimate(-m_Value, m_Error);
}

Estimate Estimate::recip() const
{
	LongFloat r(m_Value.recip());
	if (m_Error.gePow2(0)) 
		return Estimate(r, ErrorEstimate(0, ErrorEstimate::plusinf));

	if (m_Error.gePow2(-1))
		return Estimate(r, m_Error * (ErrorEstimate(1.0) - m_Error).recip());
	else 
		return Estimate(r, m_Error + ((m_Error * m_Error) << 1) + RoundingError());
}

Estimate Estimate::operator + (const Estimate &rhs) const
{
	if (m_Value.IsNegative() != rhs.m_Value.IsNegative()) return *this - (-rhs);

	LongFloat r(m_Value + rhs.m_Value);
	ErrorEstimate e(m_Error * m_Value + rhs.m_Error * rhs.m_Value);

	return Estimate(r, e * ErrorEstimate(r, ErrorEstimate::Down).recip() + RoundingError());

}

Estimate Estimate::operator - (const Estimate &rhs) const
{
	if (m_Value.IsNegative() != rhs.m_Value.IsNegative()) return *this + (-rhs);

	LongFloat r(m_Value - rhs.m_Value);
	ErrorEstimate e(m_Error * m_Value);
	ErrorEstimate e2(rhs.m_Error * rhs.m_Value);
	e = e + e2;

	if (r.Kind() == LongFloat::Zero) {
		// this is an important case
		// if substractions produces zero, offset the result by the error
		// and set the error to 2 (to cover the full interval)
		return Estimate(e.AsLongFloat(), ErrorEstimate(2.0));
	} else return Estimate(r, e * ErrorEstimate(r, ErrorEstimate::Down).recip() + RoundingError());
}

Estimate Estimate::operator * (const Estimate &rhs) const
{
	LongFloat r(m_Value * rhs.m_Value);
	ErrorEstimate e(m_Error * rhs.m_Error + m_Error + rhs.m_Error);

	return Estimate(r, e + RoundingError());
}

Estimate Estimate::operator / (const Estimate &rhs) const
{
	return *this * rhs.recip();
}

Estimate Estimate::operator << (i32 howmuch) const
{
	return Estimate(m_Value << howmuch, m_Error);
}

Estimate Estimate::operator * (i32 rhs) const
{
	LongFloat r(m_Value * rhs);

	return Estimate(r, m_Error * ErrorEstimate(double(rhs)) + RoundingError());
}

}	// namespace