/*

  RealEstimate.h

  Error math and value+error container.
  Classes:

    Estimate - combines a LongFloat with its ErrorEstimate.
		the error is relative, i.e. value * (1 - error) <
		real < value * (1 + error). If substraction results
		in a LongFloat::Zero, the result is set to the error,
		which in its turn is doubled.
	ErrorEstimate - a simple evaluation of error as 32-bit 
		mantissa and 32-bit exponent. Operations give results
		that are always greater than or equal to the actual
		result. Note the exponent in a ErrorEstimate is in 
		bits, not in words as in a LongFloat, and the mantissa
		always has 1 in its most significant bit.

*/

#ifndef FILE_REAL_ESTIMATE_H
#define FILE_REAL_ESTIMATE_H

#include <stdlib.h>
#include <limits.h>
#include "defs.h"
#include "LongFloat.h"

#define I32_MIN INT_MIN		// to be used as -inf
#define I32_MAX INT_MAX		// to be used as +inf

namespace RealLib {

class ErrorEstimate {
public:
	u32 m_Man;		// >= 2^31
	i32 m_Exp;		// error = m_Man * 2 ^ (m_Exp - 32)

	enum Inf { minusinf = I32_MIN, plusinf = I32_MAX } ;
	enum RoundingMode { Down = 0, Up = 1 };

	explicit ErrorEstimate(const u32 man = 0, const i32 exp = 0);
	// conversions (rounding up)
	ErrorEstimate(const double err);
	ErrorEstimate(const LongFloat &src, RoundingMode rnd = Up);

	ErrorEstimate RoundDownLongFloat(const LongFloat &src);

	// destructor, operator = not needed

	// operations
	// round-up addition
	ErrorEstimate operator + (const ErrorEstimate &rhs) const;
	// round-down substraction
	ErrorEstimate operator - (const ErrorEstimate &rhs) const;
	// round-up multiplication
	ErrorEstimate operator * (const ErrorEstimate &rhs) const;
	// round-up reciprocal
	ErrorEstimate recip() const;
	// round-up <<
	ErrorEstimate operator << (i32 howmuch) const;	// shift left in bits
	bool operator >= (const ErrorEstimate &rhs) const;

	ErrorEstimate& operator ++ ();	// add the minimum

	// conversions
	double AsDouble() const;
	LongFloat AsLongFloat() const;
	double DecimalsCorrect() const;

	bool gePow2 (i32 pow) const		// greater than or equal to 2^pow
	{ return m_Exp >= 1 + pow; } 

};

// two needed operations
ErrorEstimate max(const ErrorEstimate &a, const ErrorEstimate &b); 
ErrorEstimate min(const ErrorEstimate &a, const ErrorEstimate &b);

// return the rounding error introduced by current working precision
ErrorEstimate RoundingError();

class Estimate {
private:
	LongFloat m_Value;
	ErrorEstimate m_Error;		// error as m_Value * (1 +/- m_Error)

public:
	Estimate(const LongFloat &val = LongFloat(), const ErrorEstimate &err = RoundingError());
	Estimate(const Estimate &rhs)
		: m_Value(rhs.m_Value), m_Error(rhs.m_Error) {}

	// return members
	LongFloat Value() const
	{ return m_Value; }
	ErrorEstimate Error() const
	{ return m_Error; }

	// conversion to string
	char *AsDecimal(char *buffer, u32 buflen, u32 prec);

	// operations
	Estimate operator - () const;
	Estimate recip() const;
	
	Estimate operator + (const Estimate &rhs) const;
	Estimate operator - (const Estimate &rhs) const;
	Estimate operator * (const Estimate &rhs) const;
	Estimate operator / (const Estimate &rhs) const;

	// binary shift
	Estimate operator << (i32 howmuch) const;
	Estimate operator >> (i32 howmuch) const
	{ return *this << -howmuch; }

	// fast multiplication
	Estimate operator * (i32 rhs) const;

};

#define LOG_2_10 3.32192809488736234787031942948939017586483139

}	// namespace

#endif