#include "defs.h"

#include <stdlib.h>
#include <limits.h>
#include <assert.h>
#include <string.h>
#include <math.h>

#include "RealObject.h"
#include "RealFuncs.h"

#ifndef NULL
#define NULL (0)
#endif

namespace RealLib {

ObjectList *g_pEstimatesList = NULL;


RealObject::RealObject(u32 rc)
: m_RefCount(rc), m_pEstimate(NULL), m_Precision(0), m_EstimateRefs(0), m_Depth(0)
{
}

RealObject::~RealObject()
{ 
	DestroyEstimate(); 
	m_EstimateRefs = 0;
}

void RealObject::ReleaseSiblings()
{ 
}

RealObject* RealObject::AddRef() 
{ 
	assert(this); 
	++m_RefCount;

	// if there is an estimate, the new reference
	// will need one more access to it
	if (m_EstimateRefs) 
		++m_EstimateRefs;
	return this; 
}
void RealObject::Release() 
{ 
	assert(this); assert(m_RefCount > 0);

	// note: we don't decrease m_EstimateRefs
	// this is because destruction usually happens
	// after the reference has received its evaluation

	if (--m_RefCount == 0) {
		ReleaseSiblings();
		delete this; 
	}
}

// Estimate(prec):	returns an estimation of the term
// if it is referenced more than once, a copy of the
// evaluation is saved for later use

Estimate RealObject::GetEstimate(u32 prec)
{
	assert(this);
	assert(m_RefCount > 0);

	if (m_Precision != 0 && m_Precision < prec) DestroyEstimate();

	if (m_pEstimate) {
		// if this is the last time we'll need the estimate
		if (--m_EstimateRefs == 0) {		
			Estimate est = *m_pEstimate;
			DestroyEstimate();
			return est;
		} else return *m_pEstimate;
	} else if (m_RefCount > 1) {
		// create a new estimate
		m_pEstimate = new Estimate(Evaluate(prec));
		m_Precision = prec;
		m_EstimateRefs = m_RefCount - 1;

		// save it in g_pEstimatesList
		m_pPtrInObjList = new ObjectList;
		assert(m_pPtrInObjList);
		m_pPtrInObjList->next = g_pEstimatesList;
		m_pPtrInObjList->prev = NULL;
		m_pPtrInObjList->obj = this;
		if (g_pEstimatesList)
			g_pEstimatesList->prev = m_pPtrInObjList;
		g_pEstimatesList = m_pPtrInObjList;

		// return it
		return *m_pEstimate;
	} else {
		// this object is referenced only once
		// don't save, but clear m_EstimateRefs
		// which may have been set by HasEstimateReference
		m_EstimateRefs = 0;
		return Evaluate(prec);
	}
		
}

// DestroyEstimate(): if the value is cached, destroy it
// so reevaluation can begin

void RealObject::DestroyEstimate()
{ 
	if (m_pEstimate) {
		delete m_pEstimate; 
		m_pEstimate = NULL;
		m_Precision = 0;

		// delete it from g_pEstimatesList
		assert(m_pPtrInObjList);
		if (m_pPtrInObjList->prev) {
			m_pPtrInObjList->prev->next = m_pPtrInObjList->next;
		} else {
			g_pEstimatesList = m_pPtrInObjList->next;
		}
		if (m_pPtrInObjList->next) {
			m_pPtrInObjList->next->prev = m_pPtrInObjList->prev;
		}
		delete m_pPtrInObjList;

		m_pPtrInObjList = NULL;
		m_EstimateRefs = 0;
	}
}

bool RealObject::HasEstimate()
{
	return m_pEstimate != 0;
}

bool RealObject::HasEstimateReference()
{
	// return true if there is an estimate, or if the object
	// is already visited.
	if (m_EstimateRefs == 0) {
		m_EstimateRefs = 1;
		return false;
	} else return true;
}

void RealObject::SetEstimateReference()
{
	++m_EstimateRefs;
}

bool RealObject::WontBeDeleted()
{
	return m_RefCount > 1;
}

void RealObject::NonRecursiveRelease()
{
	// release the object, but do not release siblings if it
	// is deleted. They have already been visited 
	// (NonRecursiveRelease is called on GetDepthList exit,
	// from the bottom up)
	assert(this);
	assert(m_RefCount > 0);

	if (--m_RefCount == 0)
		delete this;
}

void RealObject::GetDepthList(i32 Depth, TesterFunc Test, OnExitFunc OnExit, EvalList **pList)
{
	// no siblings. only take care to call OnExit
	if (OnExit) (this->*OnExit)();
}

// real from double

RealFromDouble::RealFromDouble(const double src)
: m_Value(src)
{
}

RealFromDouble::~RealFromDouble()
{
}

Estimate RealFromDouble::Evaluate(u32 prec)
{
	return Estimate(m_Value);
}

// real from string

RealFromString::RealFromString(const char *src)
{
	// the source needs to be copied
	// otherwise the user might change it later
	assert(src);
	m_pString = new char[strlen(src) + 1];
	assert(m_pString);
	strcpy(m_pString, src);
}

RealFromString::~RealFromString()
{
	delete m_pString;
}

Estimate RealFromString::Evaluate(u32 prec)
{
	return Estimate(m_pString);
}

// real from oracle function

RealFromOracle::RealFromOracle(OracleFunction oracle)
: m_pOracle(oracle)
{
}

RealFromOracle::~RealFromOracle()
{
}

Estimate RealFromOracle::Evaluate(u32 prec)
{
	const char *val = m_pOracle(prec);
	const char *eptr = strchr(val, 'e');
	if (!eptr) eptr = strchr(val, 'E');
	if (!eptr) eptr = strchr(val, 0);	// strlen of a kind
	i32 len;
	if (strchr(val, '.')) len = eptr - val - 1;
	else len = eptr - val;
	double exp = -len * LOG_2_10;
	double man = (exp - floor(exp)) / LOG_2_10;
	ErrorEstimate er(ceil(man));
	LongFloat lf(val);
	er = er << i32(floor(exp));
	er = er * ErrorEstimate(lf.MantissaAsLongFloat()).recip();
	return Estimate(lf, er);
}

// real nullary (constants)

RealNullary::RealNullary(FuncNullary pFunc)
: m_pFunc(pFunc)
{
	assert(pFunc);
}

RealNullary::~RealNullary()
{
}

Estimate RealNullary::Evaluate(u32 prec)
{
	return m_pFunc(prec);
}

// real unary

RealUnary::RealUnary(FuncUnary pFunc, RealObject *pArg)
: m_pFunc(pFunc), m_pArg(pArg)
{
	assert(pFunc);
	assert(pArg);
	
	// reference the sibling
	pArg->AddRef();

	// mark the depth
	m_Depth = pArg->m_Depth + 1;
}

RealUnary::~RealUnary()
{
}

void RealUnary::ReleaseSiblings()
{
	m_pArg->Release();
}

// evaluate: evaluate argument, apply function to it
Estimate RealUnary::Evaluate(u32 prec)
{
	return m_pFunc(m_pArg->GetEstimate(prec), prec);
}

void RealUnary::GetDepthList(i32 Depth, TesterFunc Test, OnExitFunc OnExit, EvalList **pList)
{
	// do we need to continue?
	if (Test && (this->*Test)()) {
		if (OnExit) (this->*OnExit)();
		return;
	}

	// have we reached maximum depth?
	if (--Depth == 0) {
		if (pList) {
			EvalList *pNew = new EvalList;
			pNew->next = *pList;
			pNew->obj = this;
			pNew->used = false;
			*pList = pNew;
		}
	} else {
		// continue with sibling
		m_pArg->GetDepthList(Depth, Test, OnExit, pList);

		// clean up
		if (OnExit) (this->*OnExit)();
	}
}

static inline u32 max(u32 a, u32 b)
{ return a>b ? a : b; }

RealBinary::RealBinary(FuncBinary pFunc, RealObject *pLeft, RealObject *pRight)
: m_pFunc(pFunc), m_pLeft(pLeft), m_pRight(pRight)
{
	assert(pFunc);
	assert(pLeft);
	assert(pRight);

	// reference the siblings
	pLeft->AddRef();
	pRight->AddRef();

	// set depth (greater than both siblings)
	m_Depth = max(pLeft->m_Depth, pRight->m_Depth) + 1;
}

RealBinary::~RealBinary()
{
}

void RealBinary::ReleaseSiblings()
{
	m_pLeft->Release();
	m_pRight->Release();
}

Estimate RealBinary::Evaluate(u32 prec)
{
	return m_pFunc(m_pLeft->GetEstimate(prec), m_pRight->GetEstimate(prec), prec);
}

void RealBinary::GetDepthList(i32 Depth, TesterFunc Test, OnExitFunc OnExit, EvalList **pList)
{
	// do we need to continue?
	if (Test && (this->*Test)()) {
		if (OnExit) (this->*OnExit)();
		return;
	}

	// have we reached maximum depth?
	if (--Depth == 0) {
		if (pList) {
			EvalList *pNew = new EvalList;
			pNew->next = *pList;
			pNew->obj = this;
			pNew->used = false;
			*pList = pNew;
		}
	} else {
		// continue with the sibling with greater depth first
		// otherwise a previous run might have marked visited
		// something that could lead to a deeper dive on evaluation
		if (m_pLeft->m_Depth == m_Depth - 1) {
			m_pLeft->GetDepthList(Depth, Test, OnExit, pList);
			m_pRight->GetDepthList(Depth, Test, OnExit, pList);
		} else {
			m_pRight->GetDepthList(Depth, Test, OnExit, pList);
			m_pLeft->GetDepthList(Depth, Test, OnExit, pList);
		}

		// clean up
		if (OnExit) (this->*OnExit)();
	}
}

}  // namespace