#include "kernels.h"
#include "convolution.h"

#include <assert.h>
#include <math.h>

namespace RealLib {

i32 g_KernelsPrecision = 0;

#ifdef MULTIPLY_BY_CONVOLUTION
#define PI2 (2*3.14159265358979323846264338327950288419)
// buffers for convolution
double *g_pConvBufA = 0;
double *g_pConvBufB = 0;
// and the object itself
Convolution<double> *g_pConv = 0;
#endif

void InitializeKernels(u32 precision)
{
	g_KernelsPrecision = precision;

	#ifdef MULTIPLY_BY_CONVOLUTION
	u32 prec2pow;		
	// convolution will take 16 bits for a double
	// size should be at least twice the number of 16-bit words
	for (prec2pow = 16; prec2pow < precision*4; prec2pow <<= 1) ;
	
	g_pConvBufA = new double[prec2pow];
	g_pConvBufB = new double[prec2pow];
	g_pConv = new Convolution<double>(prec2pow, PI2);
	#endif
}

void FinalizeKernels()
{
	#ifdef MULTIPLY_BY_CONVOLUTION
	delete g_pConvBufA;
	g_pConvBufA = 0;
	delete g_pConvBufB;
	g_pConvBufB = 0;
	delete g_pConv;
	g_pConv = 0;
	#endif

	g_KernelsPrecision = 0;
}

// NormalizeMantissa: substraction can introduce zeroes in most
// significant positions of the mantissa. This function corrects
// such mantissas and returns the value that has to be substracted
// from the exponent. If this value is equal to working precision,
// the substraction function must recognize the value as Zero.
u32 NormalizeMantissa(u32 *man)
{
	i32 prec = g_KernelsPrecision - 1;
	if (man[prec] != 0) return 0;

	// find the first non-zero word
	while (--prec >= 0 && man[prec] == 0) ;
	
	// calculate needed offset
	prec = g_KernelsPrecision - (prec + 1);

	// do we have something to save?
	if (prec != g_KernelsPrecision) {
		i32 u;

		for (u=g_KernelsPrecision - 1; u >= prec; --u)
			man[u] = man[u - prec];

		for (;u >= 0; --u)
			man[u] = 0;
	}

	return prec;
}

// perform the actual addition
// returns true if there is carry
bool AddMantissa(u32 *man,			// destination, pre-initialized by a call to the default constructor
				 const u32 *full,	// the greater value
				 const u32 *part,	// the partial value,
				 i32 start)				// which is shifted by this many words
{
	int carry = 0;

	// start with carry if highest bit in what's left out is 1
	if (start != 0 && start <= g_KernelsPrecision)
		carry = part[start-1] >= (1 << 31);

	u64 v; i32 u;

	// add words
	for (u=0; u<g_KernelsPrecision - start; ++u) {
		v = u64(full[u]) + u64(part[u + start]) + u64(carry);
		man[u] = u32(v);
		carry = u32(v >> 32);
	}

	// update for carry
	for (; carry && u<g_KernelsPrecision; ++u) {
		man[u] = full[u] + carry;
		carry = man[u] == 0;
	}

	// just copy
	for (; u<g_KernelsPrecision; ++u) {
		man[u] = full[u];
	}

	return !!carry;
}

// adjust for calculations that don't fit the preallocated space.
// an extra pass might be needed if the leftover word introduces more carry.
// returns number of shifts done
u32  AdjustForCarry(u32 *man, 
					u32 msw)			// most significant word, the one that doesn't fit in
{
	// round what's left over
	u32 carry = man[0] >= (1<<31);
	i32 u;

	// shift
	for (u = 1; u<g_KernelsPrecision && carry; ++u) {
		man[u-1] = man[u] + 1;
		carry = man[u-1] == 0;
	}

	for (; u<g_KernelsPrecision; ++u)
		man[u-1] = man[u];

	// put new value
	man[u-1] = msw + carry;

	// reiterate if necessary
	if (man[u-1] == 0) return 1 + AdjustForCarry(man, 1);
	else return 1;
}

// perform the actual substraction
// returns true if part was greater and the result must be negated
bool SubMantissa(u32 *man,			// destination, pre-initialized by a call to the default constructor
				 const u32 *full,	// the greater value
				 const u32 *part,	// the partial value,
				 i32 start)				// which is shifted by this many words
{
	int carry = 0;

	// start with carry if highest bit in what's left out is 1
	if (start != 0 && start <= g_KernelsPrecision)
		carry = part[start-1] >= (1 << 31);

	u64 v; i32 u;

	// sub words
	for (u=0; u<g_KernelsPrecision - start; ++u) {
		v = u64(full[u]) - u64(part[u + start]) - u64(carry);
		man[u] = u32(v);
		carry = (v >> 32) != 0;
	}

	// update for carry
	for (; carry && u<g_KernelsPrecision; ++u) {
		man[u] = full[u] - carry;
		carry = man[u] == u32(-1);
	}

	// just copy
	for (; u<g_KernelsPrecision; ++u) {
		man[u] = full[u];
	}

	return !!carry;
}

// negate a mantissa. needed if SubMantissa returned true
void NegMantissa(u32 *man)
{
	for (i32 u = 0; u < g_KernelsPrecision; ++u)
		man[u] = ~man[u];
}

// MulMantissa
// perform actual multiplication
// the most significant word of the result is not put in man.
// instead it is returned, so no precision will be lost if
// it is zero
#ifndef MULTIPLY_BY_CONVOLUTION
u32 MulMantissa(u32 *man, 
				const u32 *a,
				const u32 *b,
				i32 inputstart,
				i32 inputlen)
#else
u32 MulMantissaDirect(u32 *man, 
				const u32 *a,
				const u32 *b,
				i32 inputstart,
				i32 inputlen)
#endif
{
	u32 carry = 0;
	u64 u, w = 0;
	i32 i = g_KernelsPrecision - inputlen * 2 + 1;
	i32 j;
	i32 k = 0;

	// start by only calculating carry
	for (; i < 0 && k < inputlen; ++i, ++k) {
		w >>= 32;
		w += u64(carry)<<32;
		carry = 0;
		for (j=0; j<=k; ++j) {
			u = u64(a[j]) * u64(b[k-j]);
			w += u;
			if (w < u) ++carry;		// this is a trick to check for carry in u64s
									// assembler would've made this a lot easier
		}
	}
	assert(i>=0);

	// we didn't write till now. 
	// besides carry, we should add 1 if the previous value had 1 in MS bit
	if (w & 0x80000000) w += u64(1)<<32;

	// start writing
	for (; k < inputlen; ++i, ++k) {
		w >>= 32;
		w += u64(carry)<<32;
		carry = 0;
		for (j=0; j<=k; ++j) {
			u = u64(a[j]) * u64(b[k-j]);
			w += u;
			if (w < u) ++carry;
		}
		man[i] = u32(w);
	}

	for (; i < g_KernelsPrecision; ++i, ++k) {
		w >>= 32;
		w += u64(carry)<<32;
		carry = 0;
		for (j=k - inputlen + 1; j<inputlen; ++j) {
			u = u64(a[j]) * u64(b[k-j]);
			w += u;
			if (w < u) ++carry;
		}
		man[i] = u32(w);
	}
	w >>= 32;
	assert(!carry);

	// leave the last word as return value
	return u32(w);
}

#ifdef MULTIPLY_BY_CONVOLUTION
u32 MulMantissa(u32 *man, 
				const u32 *a,
				const u32 *b,
				i32 inputstart,
				i32 inputlen)
{
	// do it directly if it would be faster
	if (g_KernelsPrecision < 400)		// empirical result
		return MulMantissaDirect(man, a, b, inputstart, inputlen);

	int i;
	int prec = g_KernelsPrecision;
	int prec2pow = g_pConv->GetSize();

	// initialize buffers to input
	for (i=0; i<prec; ++i) {
		g_pConvBufA[i*2] = a[i] & 0xFFFF;
		g_pConvBufA[i*2+1] = (a[i] >> 16) & 0xFFFF;
	}
	i*=2;
	while (i<prec2pow) g_pConvBufA[i++] = 0;

	for (i=inputstart; i<inputstart+inputlen; ++i) {
		g_pConvBufB[i*2] = b[i] & 0xFFFF;
		g_pConvBufB[i*2+1] = (b[i] >> 16) & 0xFFFF;
	}
	i*=2;
	while (i<prec2pow) g_pConvBufB[i++] = 0;

	// convolve
	g_pConv->Convolve(g_pConvBufA, g_pConvBufB);

	// make each value 16-bit
	double carry = 0, t;
	for (i=0; i<inputlen - 1; ++i) {
		t = floor(g_pConvBufA[i] + carry + 0.5);	// round it too
		carry = floor(ldexp(t, -16));
		g_pConvBufA[i] = t - ldexp(carry, 16);
	}
	// from here on we start writing, one in MSB of previous word is carry
	if (g_pConvBufA[i-1] > (1<<15)) carry += 1;

	for (; i<(prec + inputlen) * 2; ++i) {
		t = floor(g_pConvBufA[i] + carry + 0.5);	// round it too
		carry = floor(ldexp(t, -16));
		g_pConvBufA[i] = t - ldexp(carry, 16);
	}

	// write the result
	for (i=inputlen-1; i<prec + inputlen - 1; ++i) {
		man[i-inputlen+1] = u32(g_pConvBufA[i*2]) + (u32(g_pConvBufA[i*2+1]) << 16);
	}

	// leave the last word out
	return u32(g_pConvBufA[i*2]) + (u32(g_pConvBufA[i*2+1]) << 16);
}
#endif

// scale mantissa: multiplication by u32 multiplier
// implemented for performance
u32 ScaleMantissa(u32 *man,
				  const u32 *src,
				  u32 multiplier)
{
	u64 v = 0;

	for (i32 i=0; i<g_KernelsPrecision; ++i)
	{
		v += u64(src[i]) * u64(multiplier);
		man[i] = u32(v);
		v >>= 32;
	}

	return u32(v);
}

// binary scale mantissa, i.e. multiply by 1<<scale, where scale < 32
u32 BScaleMantissa(u32 *man,
				  const u32 *src,
				  u32 scale)
{
	u32 v = 0;

	for (i32 i=0; i<g_KernelsPrecision; ++i)
	{
		man[i] = (src[i] << scale) | v;
		v = src[i] >> (32 - scale);
	}

	return u32(v);
}

}